import 'dart:typed_data';
import 'dart:convert' as convert;

import 'package:encrypt/encrypt.dart';

import 'package:chainwallet/src/core/serialized/enums/TransactionType.dart';
import 'package:chainwallet/src/utils/Util.dart';

import 'package:pointycastle/digests/sha256.dart';
import 'package:pointycastle/digests/ripemd160.dart';

import 'ecdsa/IkeyPair.dart';
import 'ecdsa/Seed.dart';

import 'package:chainwallet/src/utils/B58IdentiferCodecs.dart';

import 'package:chainwallet/src/core/types/tx/Transaction.dart';
import 'package:chainwallet/src/core/types/signed/SignedTransaction.dart';

import 'package:chainwallet/src/model/wallet.dart';
import 'package:chainwallet/src/model/signtx.dart';

class Chainsql {
  /**
   * Generate a new account.
   *
   * @return Contains folling fields:
   * secret:Account secret.
   * address:Account address.
   * publicKey:Account publickey.
   */
  static wallet generateAddress() {
    // Security.addProvider(new BouncyCastleProvider());
    Seed seed = Seed.randomSeed();
    return generateAddressFromSeed(seed);
  }

  static wallet generateAddressFromStr(String secret) {
    Seed seed = Seed.fromBase58(secret);
    return generateAddressFromSeed(seed);
  }

  static wallet generateAddressFromSeed(Seed seed) {
    IKeyPair keyPair = seed.keyPair();
    Uint8List pubBytes = keyPair.canonicalPubBytes();

    SHA256Digest sha = new SHA256Digest();
    sha.update(pubBytes, 0, pubBytes.length);

    var result = new Uint8List(sha.digestSize);
    sha.doFinal(result, 0);

    RIPEMD160Digest d = new RIPEMD160Digest();
    d.update(result, 0, result.length);
    var o = new Uint8List(d.digestSize);
    d.doFinal(o, 0);

    String publicKey = B58IdentiferCodecs.encode(
        pubBytes, B58IdentiferCodecs.VER_ACCOUNT_PUBLIC);
    String secretKey = B58IdentiferCodecs.encodeFamilySeed(seed.bytes());
    String address = B58IdentiferCodecs.encodeAddress(o);

    return new wallet(secretKey, address, publicKey);
  }

  static Uint8List sign(Uint8List data, String secret) {
    Seed seed = Seed.fromBase58(secret);
    IKeyPair keyPair = seed.keyPair();
    return keyPair.signMessage(data);
  }

  static signtx signTx(Map txMap, String secret) {
    if (txMap == null || !txMap.containsKey('tx_json')) {
      throw Exception("missing key 'tx_json'");
    }

    Map tx = txMap['tx_json'];
    if (!tx.containsKey('TransactionType')) {
      throw Exception("missing key 'TransactionType'");
    }

    TransactionType type = TransactionType.valueOf(tx['TransactionType']);
    Transaction transaction = new Transaction(type);

    transaction.parseFromJson(tx);

    SignedTransaction signed = transaction.sign(secret);
    return new signtx(signed.tx_blob, signed.hash.toString());
  }

  static Uint8List aes256Encrypt(Uint8List plainBytes, Uint8List symKey) {
    final key = new Key(Util.paddingPass(symKey, 32));
    final iv = new IV(key.bytes.sublist(0, 16));

    String data = convert.utf8.decode(plainBytes);
    final encrypter = Encrypter(AES(key, mode: AESMode.cbc));
    final encrypted = encrypter.encrypt(data, iv: iv);
    return encrypted.bytes;
  }

  static Uint8List aes256Descrypt(Uint8List cipherBytes, Uint8List symKey) {
    final key = new Key(Util.paddingPass(symKey, 32));
    final iv = new IV(key.bytes.sublist(0, 16));
    final encrypter = Encrypter(AES(key, mode: AESMode.cbc));
    Encrypted encrypted = new Encrypted(cipherBytes);
    final decrypted = encrypter.decrypt(encrypted, iv: iv);
    return convert.utf8.encode(decrypted);
  }
}
