import 'dart:core';
import 'dart:core' as prefix0;
import "dart:typed_data";
import 'package:chainwallet/src/utils/Util.dart';

import "../serialized/SerializedType.dart";
import "../fields/Field.dart";
import "../format/Format.dart";
import "../fields/Type.dart";
import "../serialized/BytesSink.dart";
import "../serialized/BinarySerializer.dart";
import "../serialized/TypeTranslator.dart";

import 'package:chainwallet/src/core/coretypes/uint/UInt64.dart';
import 'package:chainwallet/src/core/serialized/BinaryParser.dart';
import 'package:chainwallet/src/core/coretypes/Currency.dart';
import 'package:chainwallet/src/core/coretypes/AccountID.dart';

import 'package:chainwallet/src/core/serialized/enums/TransactionType.dart';

import 'package:decimal/decimal.dart';
import 'dart:math';

class Amount implements SerializedType {
  static final int MAXIMUM_NATIVE_SCALE = 6;

  static final UInt64 BINARY_FLAG_IS_NON_NEGATIVE_NATIVE =
      new UInt64(BigInt.parse("4000000000000000", radix: 16));

  bool _isNative;

  Currency currency;

  AccountID issuer;

  Decimal value;

  UInt64 _mantissa = null;

  int _exponent;

  static AmountTranslator translate = new AmountTranslator();

  Amount(Decimal value, {String currency, String issuer}) {
    if (currency != null && issuer != null) {
      setIssuerAmount(value, currency, issuer);
    } else {
      setNativeAmount(value);
    }
  }

  void setIssuerAmount(Decimal value, String currency, String issuer) {
    _isNative = false;
    this.currency = Currency.fromString(currency);
    this.setAndCheckValue(value);

    this.issuer = AccountID.fromString(issuer);
  }

  void setNativeAmount(Decimal value) {
    _isNative = true;
    currency = Currency.ZXC;
    this.setAndCheckValue(value);
  }

  void setAndCheckValue(Decimal value) {
//    this.value = value.stripTrailingZeros();
//    initialize();

    this.value = value;
    initialize();
  }

  static int MAXIMUM_IOU_PRECISION = 16;

  int calculateExponent() {
    return -MAXIMUM_IOU_PRECISION + value.precision - value.scale;
  }

  int exponent() {
    return _exponent;
  }

  bool isNative() => _isNative;

  void initialize() {
    if (isNative()) {
      issuer = AccountID.XRP_ISSUER;
//      if (!unbounded) {
//        checkXRPBounds();
//      }
//      // Offset is unused for native amounts
      _exponent = -6; // compared to drops.
    } else {
      issuer = AccountID.NEUTRAL;
      _exponent = calculateExponent();
//
//      if (value.precision() > MAXIMUM_IOU_PRECISION && !unbounded) {
//        String err = "value precision of " + value.precision() + " is greater than maximum " +
//            "iou precision of " + MAXIMUM_IOU_PRECISION;
//        throw new PrecisionError(err, this);
//      }
    }
  }

  BigInt exactBigIntegerScaledByPowerOfTen(int n) {
    num powOfTen = pow(10, n);
    value = value * Decimal.fromInt(powOfTen);

    return BigInt.parse(value.toString());
  }

  BigInt bigIntegerDrops() {
    return exactBigIntegerScaledByPowerOfTen(MAXIMUM_NATIVE_SCALE);
  }

  UInt64 mantissa() {
    if (_mantissa == null) {
      _mantissa = calculateMantissa();
    }
    return _mantissa;
  }

  BigInt bigIntegerIOUMantissa() {
    return exactBigIntegerScaledByPowerOfTen(-_exponent).abs();
  }

  UInt64 calculateMantissa() {
    if (isNative()) {
      return new UInt64(bigIntegerDrops().abs());
    } else {
      return new UInt64(bigIntegerIOUMantissa());
    }
  }

  Map toJSON() => null;
  Uint8List toBytes() => translate.toBytes(this);
  String toHex() => null;
  void toBytesSink(BytesSink to) {
    UInt64 man = mantissa();

    if (isNative()) {
//      if (!isNegative())
//      {
      man = man.or(BINARY_FLAG_IS_NON_NEGATIVE_NATIVE);
      //}
      to.addBytes(man.toByteArray());
    } else {
      int Exponent = exponent();
      UInt64 packed;

//      if (isZero()) {
//        packed = BINARY_FLAG_IS_IOU;
//      }
//      else if (isNegative()) {
//        packed = man.or(new UInt64(512 + 0 + 97 + exponent).shiftLeft(64 - 10));
//      }
      // else {

      int iAdd = (512 + 256 + 97 + Exponent);
      BigInt bigInt = BigInt.from(iAdd);
      packed = man.or(new UInt64(bigInt).shiftLeft(64 - 10));
//    //  }
//

      // TODO issuer.bytes() ???
      to.addBytes(packed.toByteArray());
      to.addBytes(currency.bytes());
      to.addBytes(issuer.bytes());
    }
  }

  Type type() => Type.Amount;

  static void checkDropsValueWhole(String drops) {
    bool contains = drops.contains(".");
    if (contains) {
      throw new Exception("Drops string contains floating point is decimal");
    }
  }

  static Amount fromDropString(String val) {
//    BigDecimal drops = new BigDecimal(val).scaleByPowerOfTen(-6);
    // checkDropsValueWhole(val);
//    return new Amount(drops);

    Decimal drops = Decimal.parse(val) / Decimal.fromInt(1000000);
    return new Amount(drops);
  }

  static Amount fromString(String val) {
//    if (val.contains("/")) {
//      return fromIOUString(val);
//    } else if (val.contains(".")) {
//      return fromXrpString(val);
//    } else {
    return fromDropString(val);
    //}
  }

  static Decimal xrpFromDropsMantissa(Uint8List mantissa, int sign) {
    // TODO  sign?
    var bigZXC = Util.createBigIntFromUint8List(mantissa);
    return Decimal.parse(bigZXC.toString()) / Decimal.fromInt(1000000);
  }
}

class AmountTranslator extends TypeTranslator<Amount> {
//  AccountID fromParser(BinaryParser parser, int hint) {
//    return newInstance(parser.read(byteWidth()));
//  }

//  Amount fromParser(BinaryParser parser, int hint) {
//    if (hint == null) {
//      hint = 20;
//    }
//    return AccountID.fromAddressBytes(parser.read(hint));
//  }

  Amount fromString(String s) {
    return Amount.fromString(s);
  }

  Amount fromParser(BinaryParser parser, int hint) {
    Decimal value;
    Uint8List mantissa = parser.read(8);

    var sHexStr2 = Util.formatBytesAsHexString(parser.bytes);

    var sHexStr = Util.formatBytesAsHexString(mantissa);

    int b1 = mantissa[0], b2 = mantissa[1];

    bool isIOU = (b1 & 0x80) != 0;
    bool isPositive = (b1 & 0x40) != 0;
    int sign = isPositive ? 1 : -1;

    if (isIOU) {
      mantissa[0] = 0;
      Currency curr = Currency.translate.fromBinaryParser(parser);
      AccountID issuer = AccountID.translate.fromBinaryParser(parser);
      int exponent = ((b1 & 0x3F) << 2) + ((b2 & 0xff) >> 6) - 97;
      mantissa[1] &= 0x3F;

      // TODO ? 无符号的？
      var bigInt = Util.createBigIntFromUint8List(mantissa);
      value = Decimal.parse(bigInt.toString());
      num powOfTen = pow(10, -exponent);
      value = value / Decimal.fromInt(powOfTen);

      String sCurr = curr.toString();

      var test = sCurr.length;

      String sIssuer = issuer.toString();

      return new Amount(value, currency: sCurr, issuer: sIssuer);
    } else {
      mantissa[0] &= 0x3F;

      sHexStr = Util.formatBytesAsHexString(mantissa);
      value = Amount.xrpFromDropsMantissa(mantissa, sign);
      return new Amount(value);
    }
  }

  Amount fromMap(Map jsonObject) {
    String valueString = jsonObject["value"];
    String issuerString = jsonObject["issuer"];
    String currencyString = jsonObject["currency"];

    return new Amount(new Decimal.parse(valueString),
        currency: currencyString, issuer: issuerString);
  }
}
