import "dart:typed_data";

import "package:pointycastle/ecc/api.dart";
import "package:pointycastle/ecc/curves/secp256k1.dart";
import 'package:pointycastle/digests/sha512.dart';
import "package:pointycastle/export.dart";
import "package:pointycastle/api.dart";
import "package:chainwallet/src/utils/Sha512half.dart";
import "package:asn1lib/asn1lib.dart";

import "IkeyPair.dart";
import 'package:chainwallet/src/utils/Util.dart';

class K256KeyPair extends IKeyPair {
  BigInt priv, pub;
  Uint8List pubBytes;

  /**
   *
   * @param priv priv.
   * @param pub pub.
   */
  K256KeyPair(BigInt priv, BigInt pub) {
    this.priv = priv;
    this.pub = pub;
    this.pubBytes = Util.createUint8ListFromBigInt(pub);
  }

  Uint8List canonicalPubBytes() {
    return pubBytes;
  }

  BigInt getPub() {
    return pub;
  }

  BigInt getPriv() {
    return priv;
  }

  String privHex() {
    return priv.toRadixString(16);
  }

  bool verifySignature(Uint8List message, Uint8List sigBytes) {
    var parser = ASN1Parser(sigBytes);
    ASN1Sequence ss = parser.nextObject();
    ASN1Integer r2 = ss.elements[0];
    ASN1Integer s2 = ss.elements[1];

    ECSignature signature =
        new ECSignature(r2.valueAsBigInteger, s2.valueAsBigInteger);
    if (signature == null) {
      return false;
    }
    SHA512halfDigest digest = new SHA512halfDigest();
    Mac hmac = new Mac("SHA-256/HMAC");
    ECDSASigner signer = new ECDSASigner(digest, hmac);
//    ECCurve curve = new ECCurve();
//    ECPoint pubPoint = ECCurve.decodePoint(pub.toByteArray());
    ECPoint pubPoint;
    var eccDomain = new ECDomainParameters("secp256k1");
    ECPublicKey pubKey = new ECPublicKey(pubPoint, eccDomain);
    PublicKeyParameter params = new PublicKeyParameter(pubKey);
    signer.init(false, params);
    return signer.verifySignature(message, signature);
  }

  Uint8List signMessage(Uint8List message) {
    // print("原始数据\n" + Util.formatBytesAsHexString(message));

    SHA512halfDigest digest = new SHA512halfDigest();
    Mac hmac = new Mac("SHA-256/HMAC");
    ECDSASigner signer = new ECDSASigner(digest, hmac);
    var eccDomain = new ECDomainParameters("secp256k1");
    var privateKey = new ECPrivateKey(priv, eccDomain);
    var privParams = new PrivateKeyParameter(privateKey);

    signer.reset();
    signer.init(true, privParams);
    ECSignature signature = signer.generateSignature(message);
    BigInt finalS = signature.s;
    BigInt otherS = eccDomain.n - signature.s;

    if (signature.s > otherS) {
      finalS = otherS;
    }
//    if(signature.s.compareTo(otherS) == 1){
//      finalS = otherS;
//      print(finalS.toRadixString(16));
//    }

    var s = ASN1Sequence();
    s.add(ASN1Integer(signature.r));
    s.add(ASN1Integer(finalS));
    return s.encodedBytes;
  }

  static BigInt computePrivateGen(Uint8List seedBytes) {
    return generateKey(seedBytes, null);
  }

  static BigInt generateKey(Uint8List seedBytes, int discriminator) {
//    var zhongzi =  formatBytesAsHexString(seedBytes);
//    print(' 种子 :  ${zhongzi}');

    BigInt key = null;
    var keyParams = ECCurve_secp256k1();
    // keyParams.n;
    for (int i = 0; i <= 0xFFFFFFFF; i++) {
      var sha512 = new SHA512Digest();
      sha512.update(seedBytes, 0, seedBytes.length);

      if (discriminator != null) {
        sha512.updateByte((discriminator >> 24) & 0xFF);
        sha512.updateByte((discriminator >> 16) & 0xFF);
        sha512.updateByte((discriminator >> 8) & 0xFF);
        sha512.updateByte((discriminator) & 0xFF);
      }

      sha512.updateByte((i >> 24) & 0xFF);

      sha512.updateByte((i >> 16) & 0xFF);
      sha512.updateByte((i >> 8) & 0xFF);
      sha512.updateByte((i) & 0xFF);

      var out = new Uint8List(sha512.digestSize);
      ;
      sha512.doFinal(out, 0);

      var half512 = out.sublist(0, 32);
      var half512Out = Util.formatBytesAsHexString(half512);

      key = BigInt.parse(half512Out, radix: 16);

      if (key.compareTo(BigInt.zero) > 0 && key.compareTo(keyParams.n) < 0) {
        break;
      }
    }

    return key;
  }

  static Uint8List computePublicGenerator(BigInt privateGen) {
    return getPublic(privateGen);
  }

  /**
   *
   * @param privateGen privateGen
   * @param publicGenBytes publicGenBytes
   * @param accountNumber accountNumber
   * @return return value.
   */
  static BigInt computeSecretKey(
      BigInt privateGen, Uint8List publicGenBytes, int accountNumber) {
    return (generateKey(publicGenBytes, accountNumber) + privateGen) %
        ECCurve_secp256k1().n;
  }

  static Uint8List getPublic(BigInt secret) {
    var keyParams = ECCurve_secp256k1();
    ECPoint basePoint = keyParams.G;
    return (basePoint * secret).getEncoded(true);
  }

  static BigInt computePublicKey(BigInt secret) {
    var hexStr = Util.formatBytesAsHexString(getPublic(secret));
    return BigInt.parse(hexStr, radix: 16);
  }
}
